"""Implementation of operations based on index files (recursive download, recursive aaction dispatch...)."""

from __future__ import annotations

import dataclasses
import enum
import itertools
import logging
import pathlib
import typing

import requests

from . import (constants, decode, formats, json_index, path, path_directory,
               remote, task, utilities)


@dataclasses.dataclass
class IndexLoaded:
    """Message indicating that the given index file has been loaded."""

    path_id: pathlib.PurePosixPath
    """Path ID of the directory whose index has been loaded.
    """

    children: int
    """Number of subdirectories that will subsequently be loaded.
    """


@dataclasses.dataclass
class IndexProgress:
    """Represents download or process progress."""

    initial: int
    """Number of bytes already downloaded or processed when the action started.
    """

    final: int
    """Total number of bytes to download or process.
    """


@dataclasses.dataclass
class DirectoryScanned:
    """Reports information on a local directory."""

    path_id: pathlib.PurePosixPath
    """Path ID of the directory.
    """

    initial_download_count: int
    """Number of files already downloaded when the action started ("files" and "other_files").

    This count does not include -index.json.
    """

    initial_process_count: int
    """Number of files already processed when the action started ("files" and "other_files").

    This count does not include -index.json.
    """

    final_count: int  # does not include -index.json
    """Total number of files in this directory ("files" and "other_files").

    This count does not include -index.json.
    """

    index_bytes: IndexProgress
    """Size of the index file (-index.json) in bytes.
    """

    download_bytes: IndexProgress
    """Total size of the compressed files in this directory, in bytes.

    This size does not include -index.json.
    """

    process_bytes: IndexProgress
    """Total size of the files in this directory, in bytes.

    This size does not include -index.json.
    """


@dataclasses.dataclass
class Doi:
    """Message dispatched when a DOI is found in the index."""

    path_id: pathlib.PurePosixPath
    """Path ID of the associated resource.
    """

    value: str
    """Digital object identifier (DOI) string starting with ``10.``.
    """


class UncompressedDecodeProgress(task.Task):
    """Dummy task that repoorts progress on "decompression" for uncompressed resources.

    Resources that are not compressed are directly downloaded in raw format.
    The conversion from "local" to "raw" (decompression) requires no further action for such resources.
    This action dispatches decompression progress as if they were compressed, to simplify the architecture of progress trackers.
    """

    def __init__(self, path_id: pathlib.PurePosixPath, size: int):
        self.path_id = path_id
        self.size = size

    def run(self, session: requests.Session, manager: task.Manager):
        manager.send_message(
            decode.Progress(
                path_id=self.path_id,
                initial_bytes=0,
                current_bytes=self.size,
                final_bytes=0,
                complete=True,
            )
        )


class Selector:
    """Delegate called to pick an action for each file.

    Selectors are used during the indexing phase to calculate the number of bytes to download and/or process,
    and during the processing phase to choose the action to perform.
    """

    class Action(enum.Enum):
        """Specifies the operation to perform for a given file.

        The action also determines whether the file's bytes should be
        accounted for during the indexing phase.
        This is useful to report non-zero progress after resuming a job,
        but skip the actual processing.
        """

        IGNORE = 0
        """Skips this file and does not report it.
        """
        DOI = 1
        """Skips this file, does not report it, but publishes own DOIs.
        """
        SKIP = 2
        """Skips this file but reports it as downloaded and processed.
        """

        DOWNLOAD_SKIP = 3
        """Skips operations on this file but reports it as downloaded.
        """
        DOWNLOAD = 4
        """Downloads and reports.
        """
        DECOMPRESS = 5
        """Downloads, decompresses, and reports.
        """
        PROCESS = 6
        """Downloads, decompresses, processes, and reports.
        """

    SKIP_ACTIONS = {Action.SKIP, Action.DOWNLOAD_SKIP}
    """The set of actions that skip all operations on the file.
    """

    REPORT_DOWNLOAD_ACTIONS = {
        Action.SKIP,
        Action.DOWNLOAD_SKIP,
        Action.DOWNLOAD,
        Action.DECOMPRESS,
        Action.PROCESS,
    }
    """The set of actions that (at least) download the file.
    """

    REPORT_PROCESS_ACTIONS = {Action.SKIP, Action.DECOMPRESS, Action.PROCESS}
    """The set of actions that download and process the file.
    """

    INSTALL_IGNORE_ACTIONS = {
        Action.IGNORE,
        Action.DOI,
        Action.SKIP,
        Action.DOWNLOAD_SKIP,
    }
    """The set of actions that ignore the file for reporting purposes.
    """

    def action(self, file: path.File) -> "Selector.Action":
        """Returns the action to apply to the given file.

        Called by :py:class:`Index`, :py:class:`InstallFilesRecursive` and :py:class:`ProcessFilesRecursive`.
        The default implementation returns `Selector.Action.PROCESS`.
        """
        return Selector.Action.PROCESS

    def scan_filesystem(self, directory: path_directory.Directory) -> bool:
        """Whether to scan the filesystem.

        Called by :py:class:`Index` to decide whether it needs to scan the file system.
        This function may return False if :py:func:`action` returns one of the following for every file in the directory:

        - :py:attr:`Selector.Action.IGNORE`
        - :py:attr:`Selector.Action.DOI`
        - :py:attr:`Selector.Action.SKIP`
        - :py:attr:`Selector.Action.DOWNLOAD_SKIP`
        """
        return True


class Index(remote.DownloadFile):
    """Downloads an index file (-index.json).

    Args:
        path_root (pathlib.Path): The root path used to generate local file paths.
        path_id (pathlib.PurePosixPath): The path ID of the directory that will be seached recursively.
        server (remote.Server): The remote server to download resources.
        selector (Selector): A selector that defines the files to process.
        priority (int): Priority of this task (tasks with lower priorities are scheduled first).
        force (bool): Download the index file even if it is already present locally.
        directory_doi (bool): Whether to dispatch :py:class:`Doi` messages while reading the index.
    """

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        server: remote.Server,
        selector: Selector,
        priority: int,
        force: bool,
        directory_doi: bool,
    ):
        self.index_file = path.File(
            path_root=path_root,
            path_id=path_id / "-index.json",
            own_doi=None,
            metadata={},
            server=server,
            # size, hash, and compressions are not available for index files
            # they are always non-null when path.File objects are exposed to the user via Selector,
            # so we keep the type as a non-optional
            size=None,  # type: ignore
            hash=None,  # type: ignore
            compressions=None,  # type: ignore
            session=None,
            manager=task.NullManager(),
        )
        super().__init__(
            path_root=path_root,
            path_id=path_id / "-index.json",
            suffix=None,
            server=server,
            force=force,
            expected_size=None,
            expected_hash=None,
        )
        self.selector = selector
        self.priority = priority
        self.directory_doi = directory_doi

    def run(self, session: requests.Session, manager: task.Manager):
        directory = path_directory.Directory(
            path_root=self.path_root,
            path_id=self.path_id.parent,
            own_doi=None,
            metadata={},
            server=self.server,
            doi_and_metadata_loaded=False,
        )
        directory.local_path.mkdir(exist_ok=True)
        directory_scanned = DirectoryScanned(
            path_id=directory.path_id,
            initial_download_count=0,
            initial_process_count=0,
            final_count=0,
            index_bytes=IndexProgress(initial=0, final=0),
            download_bytes=IndexProgress(initial=0, final=0),
            process_bytes=IndexProgress(initial=0, final=0),
        )
        if not self.force:
            if self.index_file.local_path.is_file():
                directory_scanned.index_bytes.initial = (
                    self.index_file.local_path.stat().st_size
                )
            else:
                partial_index_path = utilities.path_with_suffix(
                    self.path_root / self.path_id, constants.DOWNLOAD_SUFFIX
                )
                if partial_index_path.is_file():
                    directory_scanned.index_bytes.initial = (
                        partial_index_path.stat().st_size
                    )
        super().run(session=session, manager=manager)
        index_data = json_index.load(self.path_root / self.path_id)
        if "doi" in index_data:
            directory.__dict__["own_doi"] = index_data["doi"]
        if "metadata" in index_data:
            directory.__dict__["metadata"] = index_data["metadata"]
        manager.send_message(
            IndexLoaded(
                path_id=directory.path_id,
                children=len(index_data["directories"]),
            )
        )
        for child_directory_name in index_data["directories"]:
            manager.schedule(
                Index(
                    path_root=self.path_root,
                    path_id=directory.path_id / child_directory_name,
                    server=self.server,
                    selector=self.selector,
                    priority=self.priority,
                    force=self.force,
                    directory_doi=self.directory_doi,
                ),
                priority=self.priority,
            )
        if self.selector.scan_filesystem(directory=directory):
            name_to_size = {
                path.name: path.stat().st_size
                for path in directory.local_path.iterdir()
            }
        else:
            name_to_size = {"-index.json": self.index_file.local_path.stat().st_size}
        directory_scanned.index_bytes.final = name_to_size["-index.json"]
        if self.directory_doi:
            if directory.own_doi is not None:
                manager.send_message(
                    Doi(path_id=directory.path_id, value=directory.own_doi)
                )
        for file in itertools.chain(
            (
                formats.file_from_dict(data=data, parent=directory)
                for data in index_data["files"]
            ),
            (
                path.File.from_dict(data=data, parent=directory)
                for data in index_data["other_files"]
            ),
        ):
            action = self.selector.action(file)
            if action == Selector.Action.IGNORE:
                continue
            if action == Selector.Action.DOI:
                if file.own_doi is not None:
                    manager.send_message(Doi(path_id=file.path_id, value=file.own_doi))
                continue
            directory_scanned.final_count += 1
            if action in Selector.REPORT_DOWNLOAD_ACTIONS:
                directory_scanned.download_bytes.final += file.best_compression.size
                if action in Selector.REPORT_PROCESS_ACTIONS:
                    directory_scanned.process_bytes.final += file.size
                if action in Selector.SKIP_ACTIONS:
                    directory_scanned.initial_download_count += 1
                    directory_scanned.download_bytes.initial += (
                        file.best_compression.size
                    )
                    if action in Selector.REPORT_PROCESS_ACTIONS:
                        directory_scanned.initial_process_count += 1
                        directory_scanned.process_bytes.initial += file.size
                elif not self.force:
                    if file.path_id.name in name_to_size:
                        directory_scanned.initial_download_count += 1
                        directory_scanned.download_bytes.initial += (
                            file.best_compression.size
                        )
                        if action in Selector.REPORT_PROCESS_ACTIONS:
                            directory_scanned.initial_process_count += 1
                            directory_scanned.process_bytes.initial += file.size
                    else:
                        compressed_name = (
                            f"{file.path_id.name}{file.best_compression.suffix}"
                        )
                        if compressed_name in name_to_size:
                            directory_scanned.initial_download_count += 1
                            directory_scanned.download_bytes.initial += (
                                file.best_compression.size
                            )
                        else:
                            partial_compressed_name = (
                                f"{compressed_name}{constants.DOWNLOAD_SUFFIX}"
                            )
                            if (
                                # in process mode, files are not persisted to the disk
                                # data is downloaded (or read from the disk), decompressed if necessaary,
                                # and processed in chunks
                                # for simplicity, partially persisted downloads are ignored and re-downloaded from scratch
                                action != Selector.Action.PROCESS
                                and partial_compressed_name in name_to_size
                            ):
                                directory_scanned.download_bytes.initial += (
                                    name_to_size[partial_compressed_name]
                                )
        manager.send_message(directory_scanned)


class InstallFilesRecursive(task.Task):
    """Downloads (and possibly decompresses) a directories' files recursively.

    The actual action is controlledd by the selector aand may be different for different files. Child directories are installed recursively.

    Args:
        path_root (pathlib.Path): The root path used to generate local file paths.
        path_id (pathlib.PurePosixPath): The path ID of the directory that will be seached recursively.
        server (remote.Server): The remote server to download resources.
        selector (Selector): A selector that defines the files to process.
        priority (int): Priority of this task (tasks with lower priorities are scheduled first).
        force (bool): Download files even if they already present locally.
    """

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        server: remote.Server,
        selector: Selector,
        priority: int,
        force: bool,
    ):
        self.path_root = path_root
        self.path_id = path_id
        self.server = server
        self.selector = selector
        self.priority = priority
        self.force = force

    def run(self, session: requests.Session, manager: task.Manager):
        directory = path_directory.Directory(
            path_root=self.path_root,
            path_id=self.path_id,
            own_doi=None,
            metadata={},
            server=self.server,
            doi_and_metadata_loaded=False,
        )
        index_data = json_index.load(directory.local_path / "-index.json")
        if self.selector.scan_filesystem(directory=directory):
            names = set(path.name for path in directory.local_path.iterdir())
        else:
            names: set[str] = set()
        for file in itertools.chain(
            (
                formats.file_from_dict(data=data, parent=directory)
                for data in index_data["files"]
            ),
            (
                path.File.from_dict(data=data, parent=directory)
                for data in index_data["other_files"]
            ),
        ):
            file.attach_session(session)
            file.attach_manager(manager)
            action = self.selector.action(file)
            if action in Selector.INSTALL_IGNORE_ACTIONS:
                continue
            if (
                not self.force
                and action == Selector.Action.PROCESS
                and not file.path_id.name in names
                and not f"{file.path_id.name}{file.best_compression.suffix}" in names
                and f"{file.path_id.name}{file.best_compression.suffix}{constants.DOWNLOAD_SUFFIX}"
                in names
            ):
                utilities.path_with_suffix(
                    file.local_path,
                    f"{file.best_compression.suffix}{constants.DOWNLOAD_SUFFIX}",
                ).unlink()
            # 0: skip
            # 1: download
            # 2: download and publish decompress progress when done (uncompressed file)
            # 3: download and decompress
            # 4: decompress
            actual_action: int = 0
            if self.force:
                if action == Selector.Action.DOWNLOAD:
                    actual_action = 1
                elif isinstance(file.best_compression, decode.NoneCompression):
                    actual_action = 2
                else:
                    actual_action = 3
            elif file.path_id.name in names:
                actual_action = 0
            elif action == Selector.Action.DOWNLOAD:
                if f"{file.path_id.name}{file.best_compression.suffix}" in names:
                    actual_action = 0
                else:
                    actual_action = 1
            elif isinstance(file.best_compression, decode.NoneCompression):
                actual_action = 2
            else:
                if f"{file.path_id.name}{file.best_compression.suffix}" in names:
                    actual_action = 4
                else:
                    actual_action = 3
            logging.debug(
                f"path_id={file.path_id} force={self.force} {action=} {actual_action=}"
            )
            if actual_action != 0:
                download_task = remote.DownloadFile(
                    path_root=self.path_root,
                    path_id=file.path_id,
                    suffix=file.best_compression.suffix,
                    server=self.server,
                    force=self.force,
                    expected_size=file.best_compression.size,
                    expected_hash=file.best_compression.hash,
                )
                decompress_task = decode.DecompressFile(
                    path_root=self.path_root,
                    path_id=file.path_id,
                    compression=file.best_compression,
                    expected_size=file.size,
                    expected_hash=file.hash,
                    word_size=file.word_size,
                    keep=False,
                )
                if actual_action == 1:
                    manager.schedule(download_task, self.priority)
                elif actual_action == 2:
                    manager.schedule(
                        task.Chain(
                            (
                                download_task,
                                UncompressedDecodeProgress(
                                    path_id=file.path_id,
                                    size=file.size,
                                ),
                            )
                        ),
                        self.priority,
                    )
                elif actual_action == 3:
                    manager.schedule(
                        task.Chain((download_task, decompress_task)), self.priority
                    )
                elif actual_action == 4:
                    manager.schedule(decompress_task, self.priority)
                else:
                    raise Exception(f"unexpected action {actual_action}")
        for child_directory_name in index_data["directories"]:
            manager.schedule(
                InstallFilesRecursive(
                    path_root=self.path_root,
                    path_id=self.path_id / child_directory_name,
                    server=self.server,
                    selector=self.selector,
                    priority=self.priority,
                    force=self.force,
                ),
                priority=self.priority,
            )


class ProcessFile(task.Task):
    """Generic task for file processing.

    Args:
        file (path.File): The file (remote or local) to process.
    """

    def __init__(self, file: path.File):
        self.file = file


ProcessFileType = typing.TypeVar("ProcessFileType", bound=ProcessFile)
"""Generic parameter representing the file type.

Used by :py:class:`ProcessFilesRecursive`.
"""


class ProcessFilesRecursive(task.Task):
    """Spawns a processing task for each file in the given directory.

    Subdirectories are recursively searched as well.

    Args:
        path_root (pathlib.Path): The root path used to generate local file paths.
        path_id (pathlib.PurePosixPath): The path ID of the directory that will be scanned recursively.
        server (remote.Server): The remote server used to download resources.
        selector (Selector): A selector that defines the files to process.
        process_file_class (typing.Type[ProcessFileType]): The class of the task to run on each selected file. Must be a subclass of :py:class:`ProcessFile`.
        process_file_args (typing.Iterable[typing.Any]): Positional arguments passed to the constructor of `process_file_class`.
        process_file_kwargs (typing.Mapping[str, typing.Any]): Keyword arguments passed to the constructor of `process_file_class`. The keyword argument `file` is automatically added by `ProcessFilesRecursive` after the positional arguments and before other keyword arguments.
        priority (int): Priority of this task and all recursively created tasks (tasks with lower priorities are scheduled first).
    """

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        server: remote.Server,
        selector: Selector,
        process_file_class: typing.Type[ProcessFileType],
        process_file_args: typing.Iterable[typing.Any],
        process_file_kwargs: typing.Mapping[str, typing.Any],
        priority: int,
    ):
        self.path_root = path_root
        self.path_id = path_id
        self.server = server
        self.process_file_class = process_file_class
        self.process_file_args = process_file_args
        self.process_file_kwargs = process_file_kwargs
        self.selector = selector
        self.priority = priority

    def run(self, session: requests.Session, manager: task.Manager) -> None:
        directory = path_directory.Directory(
            path_root=self.path_root,
            path_id=self.path_id,
            own_doi=None,
            metadata={},
            server=self.server,
            doi_and_metadata_loaded=False,
        )
        index_data = json_index.load(directory.local_path / "-index.json")
        for file in itertools.chain(
            (
                formats.file_from_dict(data=data, parent=directory)
                for data in index_data["files"]
            ),
            (
                path.File.from_dict(data=data, parent=directory)
                for data in index_data["other_files"]
            ),
        ):
            action = self.selector.action(file)
            if action == Selector.Action.PROCESS:
                manager.schedule(
                    task=self.process_file_class(
                        *self.process_file_args,
                        file=file,
                        **self.process_file_kwargs,
                    ),
                    priority=self.priority,
                )
        for child_directory_name in index_data["directories"]:
            manager.schedule(
                ProcessFilesRecursive(
                    path_root=self.path_root,
                    path_id=self.path_id / child_directory_name,
                    server=self.server,
                    selector=self.selector,
                    process_file_class=self.process_file_class,
                    process_file_args=self.process_file_args,
                    process_file_kwargs=self.process_file_kwargs,
                    priority=self.priority,
                ),
                priority=self.priority,
            )



"""Basic index files parsing."""

from __future__ import annotations

import errno
import functools
import json
import pathlib
import typing

from . import constants, utilities

schema = utilities.load_schema("-index_schema.json")
"""JSON schema for -index files."""


class InstallError(FileNotFoundError):
    """Raised if the target path does not exist.

    Args:
        path (pathlib.Path): The path that does not exist.
    """

    def __init__(self, path: pathlib.Path):
        super().__init__(errno.ENOENT, "", str(path))

    def __str__(self) -> str:
        return (
            f'"{self.filename}" does not exist, have you installed the configuration?'
        )


@functools.lru_cache(maxsize=constants.LRU_CACHE_MAXSIZE)
def load(path: pathlib.Path) -> dict[str, typing.Any]:
    """Reads and validates a -index.json file.

    This function caches the parsed contents of up to :py:attr:`undr.constants.LRU_CACHE_MAXSIZE` files.

    Args:
        path (pathlib.Path): The path of the file to read.

    Raises:
        InstallError: if the file does not exist.
        jsonschema_rs.ValidationError: if validation fails.

    Returns:
        dict[str, typing.Any]: Parsed JSON file contents.
    """

    try:
        with open(path) as index_data_file:
            index_data = json.load(index_data_file)
        schema.validate(index_data)
        return index_data
    except FileNotFoundError:
        raise InstallError(path)



"""Implementation of all supported compression formats.

To add support for a new format, create and implement a derived class of :py:class:`Compression` (see for instance :py:class:`BrotliCompression`) and add it to :py:func:`compression_from_dict.`.
"""

from __future__ import annotations

import dataclasses
import pathlib
import typing

import requests

from . import constants, exception, task, utilities


@dataclasses.dataclass
class Progress:
    """Represents decompression progress for a given resource."""

    path_id: pathlib.PurePosixPath
    """Identifier of the resource.
    """

    initial_bytes: int
    """Number of bytes of the decompressed resource that were already decompressed when the current decoding process began.
    """

    current_bytes: int
    """Number of bytes of the decompressed resource that have been decompressed so far.
    """

    final_bytes: int
    """Total number of bytes of the decompressed resource.
    """

    complete: bool
    """Whether decoding of this resource is complete.
    """


class RemainingBytesError(Exception):
    """Raised if the number of bytes in the decompressed resource is not a multiple of its word size.

    Args:
        word_size (int): The resource's word size.
        buffer (bytes): The remaining bytes. Their length is larger than zero and smaller than the word size.
    """

    def __init__(self, word_size: int, buffer: bytes):
        self.buffer = buffer
        super().__init__(
            f"The total number of bytes is not a multiple of {word_size} ({len(buffer)} remaining)"
        )


class Decoder:
    """Abstract class for decoders. A decoder controls a decompression process."""

    def decompress(self, buffer: bytes) -> bytes:
        """Consumes a buffer and produces decompressed bytes.

        Args:
            buffer (bytes): Compressed input bytes.

        Returns:
            bytes: Decompressed output bytes. Their length must be a multiple of the word size.
        """
        raise NotImplementedError()

    def finish(self) -> tuple[bytes, bytes]:
        """Tells the decoder that all input bytes have been read.

        Returns:
            tuple[bytes, bytes]: Decompressed output bytes, whose length must be a multiple of the word size, and remaining bytes, whose length must be striclty smaller than the word size. A non-zero number of remaining bytes usually indicates an issue (erroneous configuration or corrupted data).
        """
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class Compression:
    """Represents a compressed file's metadata."""

    suffix: str
    """Suffix for files compressed with this format.

    The suffix must include a leading dot, for instance ``".br"``.
    """

    size: int
    """Size of the compressed file in bytes.
    """

    hash: str
    """SHA3-224 (FIPS 202) hash of the compressed bytes.
    """

    def decoder(self, word_size: int) -> Decoder:
        """Creates a new decoder for this compression.

        Args:
            word_size (int): The resource's word size in bytes.

        Returns:
            Decoder: A decompression manager for this compression format.
        """
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class NoneCompression(Compression):
    """Placeholder format for uncompressed files.

    This "compression" format passes the input bytes to the output without transforming them.
    It may cut and stitch buffers to ensure that each buffer has a length that is a multiple of the word size.

    Args:
        word_size (int): The resource's word size in bytes.
    """

    class Decoder(Decoder):
        def __init__(self, word_size: int):
            assert word_size > 0
            self.word_size = word_size
            self.buffer = bytes()

        def decompress(self, buffer: bytes):
            if len(self.buffer) == 0:
                if len(buffer) % self.word_size == 0:
                    return buffer
                else:
                    remaning_bytes_count = len(buffer) % self.word_size
                    buffer, self.buffer = (
                        buffer[0 : len(self.buffer) - remaning_bytes_count],
                        buffer[len(buffer) - remaning_bytes_count :],
                    )
                    return buffer
            remaning_bytes_count = (len(self.buffer) + len(buffer)) % self.word_size
            if remaning_bytes_count == 0:
                buffer, self.buffer = self.buffer + buffer, bytes()
                return buffer
            if remaning_bytes_count < len(buffer):
                buffer, self.buffer = (
                    self.buffer + buffer[0 : len(buffer) - remaning_bytes_count],
                    buffer[len(buffer) - remaning_bytes_count :],
                )
                return buffer
            if remaning_bytes_count == len(buffer):
                buffer, self.buffer = self.buffer, buffer
                return buffer
            buffer, self.buffer = (
                self.buffer[0 : len(self.buffer) + len(buffer) - remaning_bytes_count],
                self.buffer[len(self.buffer) + len(buffer) - remaning_bytes_count :]
                + buffer,
            )
            return buffer

        def finish(self):
            remaning_bytes_count = len(self.buffer) % self.word_size
            return (
                self.buffer[0 : len(self.buffer) - remaning_bytes_count],
                self.buffer[len(self.buffer) - remaning_bytes_count :],
            )

    def decoder(self, word_size: int):
        return NoneCompression.Decoder(word_size=word_size)


@dataclasses.dataclass(frozen=True)
class BrotliCompression(Compression):
    """Implements Brotli decompression (https://github.com/google/brotli).

    Args:
        word_size (int): The resource's word size in bytes.
    """

    class Decoder(NoneCompression.Decoder):
        def __init__(self, word_size: int):
            super().__init__(word_size=word_size)
            import brotli

            self.decoder = brotli.Decompressor()

        def decompress(self, buffer: bytes):
            decompressed_bytes = self.decoder.process(buffer)

            return super().decompress(decompressed_bytes)

        def finish(self):
            if not self.decoder.is_finished():
                raise Exception("the Brotli decoder expected more data")
            return super().finish()

    def decoder(self, word_size: int):
        return BrotliCompression.Decoder(word_size=word_size)


class DecompressFile(task.Task):
    """Decompresses a local file and writes decoded bytes to another local file.

    Args:
        path_root (pathlib.Path): The root path used to generate local file paths.
        path_id (pathlib.PurePosixPath): The path ID of the file.
        compression (Compression): The format of the compressed file.
        expected_size (int): The size of the decompressed file in bytes, according to the index.
        expected_hash (str): The hash of the decompressed file, according to the index.
        word_size (int): The file's word size (the number of decoded bytes must be a multiple of this value).
        keep (bool): Whether to keep the compressed file after a successful decompression.
    """

    def __init__(
        self,
        path_root: pathlib.Path,
        path_id: pathlib.PurePosixPath,
        compression: Compression,
        expected_size: int,
        expected_hash: str,
        word_size: int,
        keep: bool,
    ):
        self.path_root = path_root
        self.path_id = path_id
        self.compression = compression
        self.expected_size = expected_size
        self.expected_hash = expected_hash
        self.word_size = word_size
        self.keep = keep

    def run(self, session: requests.Session, manager: task.Manager):
        hash = utilities.new_hash()
        decompress_path = utilities.path_with_suffix(
            self.path_root / self.path_id, constants.DECOMPRESS_SUFFIX
        )
        with open(
            utilities.path_with_suffix(
                self.path_root / self.path_id, self.compression.suffix
            ),
            "rb",
        ) as compressed_file:
            with open(decompress_path, "wb") as decompressed_file:
                decoder = self.compression.decoder(word_size=self.word_size)
                while True:
                    compressed_buffer = compressed_file.read(constants.CHUNK_SIZE)
                    if len(compressed_buffer) == 0:
                        break
                    decompressed_buffer = decoder.decompress(compressed_buffer)
                    decompressed_file.write(decompressed_buffer)
                    hash.update(decompressed_buffer)
                    manager.send_message(
                        Progress(
                            path_id=self.path_id,
                            initial_bytes=0,
                            current_bytes=len(decompressed_buffer),
                            final_bytes=len(decompressed_buffer),
                            complete=False,
                        )
                    )
                (decompressed_buffer, remaining_bytes) = decoder.finish()
                decompressed_file.write(decompressed_buffer)
                hash.update(decompressed_buffer)
                manager.send_message(
                    Progress(
                        path_id=self.path_id,
                        initial_bytes=0,
                        current_bytes=len(decompressed_buffer),
                        final_bytes=len(decompressed_buffer),
                        complete=False,
                    )
                )
                if len(remaining_bytes) > 0:
                    raise RemainingBytesError(
                        word_size=self.word_size, buffer=remaining_bytes
                    )
        size = decompress_path.stat().st_size
        if size != self.expected_size:
            raise exception.SizeMismatch(self.path_id, self.expected_size, size)
        digest = hash.hexdigest()
        if digest != self.expected_hash:
            exception.HashMismatch(self.path_id, self.expected_hash, digest)
        decompress_path.replace(self.path_root / self.path_id)
        if not self.keep:
            utilities.path_with_suffix(
                self.path_root / self.path_id, self.compression.suffix
            ).unlink()
        manager.send_message(
            Progress(
                path_id=self.path_id,
                initial_bytes=0,
                current_bytes=0,
                final_bytes=0,
                complete=True,
            )
        )


def compression_from_dict(
    data: dict[str, typing.Any], base_size: int, base_hash: str
) -> Compression:
    """Factory for comprssion formats.

    Args:
        data (dict[str, typing.Any]): Parsed compression object read from an index file.
        base_size (int): Size of the uncompressed file in bytes, read from the index.
        base_hash (str): Hash of the uncompressed file, read from the index.

    Raises:
        RuntimeError: if the compression format is unknown or not supported.

    Returns:
        Compression: The compressed file's metadata, can be used to create a decoder.
    """
    if data["type"] == "none":
        return NoneCompression(suffix=data["suffix"], size=base_size, hash=base_hash)
    if data["type"] == "brotli":
        return BrotliCompression(
            suffix=data["suffix"], size=data["size"], hash=data["hash"]
        )
    raise RuntimeError("unsupported compression type {}".format(data["type"]))



"""Data loading and number formatting functions."""

from __future__ import annotations

import hashlib
import json
import math
import pathlib
import pkgutil
import typing

import jsonschema_rs

from . import constants


def load_schema(name: str) -> jsonschema_rs.JSONSchema:
    """Reads and parses a JSON schema bundled with UNDR.

    Args:
        name (str): Name of the schema.

    Returns:
        jsonschema_rs.JSONSchema: JSON schema validator.
    """
    data = pkgutil.get_data("undr", name)
    assert data is not None
    return jsonschema_rs.JSONSchema(json.loads(data.decode()))


def least_multiple_over_chunk_size(word_size: int) -> int:
    """Calculates the maximum number of bytes in a chunk that can be divided into full words.

    For instance, for a chunks size of 100 bytes and a word size of 32 bytes, this function would return 96.

    Args:
        word_size (int): The word size in bytes. The chunk size is not a parameter since UNDR always uses :py:attr:`undr.constants.CHUNK_SIZE`.

    Returns:
        int: Maximum number of bytes in a chunk that can be divided into full words. This number is guaranteed to be a multiple of word_size. It may be zero.
    """
    return word_size * int(math.ceil(constants.CHUNK_SIZE / word_size))


def path_with_suffix(path: pathlib.Path, suffix: str) -> pathlib.Path:
    """Appends a suffix to a path and returns a new path.

    Args:
        path (pathlib.Path): Input path, not modified by this function.
        suffix (str): The string to append to the path's last component.

    Returns:
        pathlib.Path: New path with the given suffix.
    """
    return pathlib.Path(f"{path}{suffix}")


def posix_path_with_suffix(
    path: pathlib.PurePosixPath, suffix: str
) -> pathlib.PurePosixPath:
    """Appends a suffix to a POSIX path and returns a new path.

    Similar to :py:func:`path_with_suffix` for POSIX paths.

    Args:
        path (pathlib.PurePosixPath): Input path, not modified by this function.
        suffix (str): The string to append to the path's last component.

    Returns:
        pathlib.PurePosixPath: New path with the given suffix.
    """
    return pathlib.PurePosixPath(f"{path}{suffix}")


def new_hash() -> "hashlib._Hash":
    """Creates a new byte hasher.

    Returns:
        hashlib._Hash: SHA3-224 (FIPS 202) hasher.
    """
    return hashlib.sha3_224()


def hash(chunks: typing.Iterable[bytes]) -> "hashlib._Hash":
    """Consumes an iterable and calculates a hash.

    Since this function consumes the hash, users should use :py:func:`new_hash` and call :py:meth:`hashlib._Hash.update` manually if they plan to do something else with the bytes.

    Args:
        chunks (typing.Iterable[bytes]): A bytes iterable.

    Returns:
        hashlib._Hash: SHA3-224 (FIPS 202) hasher. Use :py:meth:`hashlib._Hash.digest` or :py:meth:`hashlib._Hash.hexdigest` to read the hash value.
    """
    hash_object = new_hash()
    for chunk in chunks:
        hash_object.update(chunk)
    return hash_object


def hash_file(path: pathlib.Path, chunk_size: int):
    """Calculates a file's hash.

    Args:
        path (pathlib.Path): Path of the file to hash.
        chunk_size (int): Chunk size in bytes, used to read the file.

    Returns:
        _type_: _description_
    """
    with open(path, "rb") as input:
        return hash(iter(lambda: input.read(chunk_size), b""))


def duration_to_string(duration: float) -> str:
    """Generates a human-readable representation of a duration.

    Args:
        duration (float): Positive time delta in seconds.

    Returns:
        str: Human-redable representation.
    """
    duration = round(duration)
    if duration < 180:
        return f'{"{:.0f}".format(duration)} s'
    if duration < 10800:
        return f'{"{:.0f}".format(math.floor(duration / 60))} min'
    if duration < 259200:
        return f'{"{:.0f}".format(math.floor(duration / 3600))} h'
    return f'{"{:.0f}".format(math.floor(duration / 86400))} days'


def size_to_string(size: int) -> str:
    """Generates a human-readable representation of a size.

    Args:
        size (float): Resource size in bytes.

    Returns:
        str: Human-redable representation.
    """
    if size < 1000:
        return f'{"{:.0f}".format(size)} B'
    if size < 1000000:
        return f'{"{:.2f}".format(size / 1000)} kB'
    if size < 1000000000:
        return f'{"{:.2f}".format(size / 1000000)} MB'
    if size < 1000000000000:
        return f'{"{:.2f}".format(size / 1000000000)} GB'
    return f'{"{:.2f}".format(size / 1000000000000)} TB'


def speed_to_string(speed: int) -> str:
    """Generates a human-readable representation of a speed.

    Args:
        speed (int): Download or process speed in bytes per second.

    Returns:
        str: Human-redable representation.
    """
    if speed < 1000:
        return f'{"{:.0f}".format(speed)} B/s'
    if speed < 1000000:
        return f'{"{:.2f}".format(speed / 1000)} kB/s'
    if speed < 1000000000:
        return f'{"{:.2f}".format(speed / 1000000)} MB/s'
    if speed < 1000000000000:
        return f'{"{:.2f}".format(speed / 1000000000)} GB/s'
    return f'{"{:.2f}".format(speed / 1000000000000)} TB/s'



"""Progress storage utility."""

from __future__ import annotations

import collections
import dataclasses
import os
import pathlib
import sqlite3
import threading
import time
import types
import typing


@dataclasses.dataclass
class Progress:
    """Message that indicates that the given resource has been persisted."""

    path_id: pathlib.PurePosixPath
    """The resource's unique path ID.
    """


class ReadOnlyStore:
    """Stores the IDs of processed tasks.

    This store provides a method to check whether a task has been performed but it cannot be modified.
    Most users will probably prefer the writable :py:class:`Store`.

    Args:
        path (typing.Union[str, os.PathLike]): Path of the SQLite database file with extension ".db".
    """

    def __init__(
        self,
        path: typing.Union[str, os.PathLike],
    ):
        self.path = pathlib.Path(path).resolve()
        self.connection = sqlite3.connect(self.path)
        self.cursor = self.connection.cursor()

        rows = [row for row in self.cursor.execute("pragma table_info(complete)")]
        if len(rows) == 0:
            self.cursor.execute(
                "create table complete (id text primary key) without rowid"
            )
            self.connection.commit()
        elif len(rows) != 1 or rows[0] != (0, "id", "TEXT", 1, None, 1):
            raise Exception('the table "complete" does not have the expected format')

    def __contains__(self, id: str):
        """Whether the given ID has been processed.

        Args:
            id (str): The ID to check.

        Returns:
            bool: True if the file is in the store, which means that it has been processed.
        """
        return (
            self.cursor.execute(
                "select * from complete where id == ?", (id,)
            ).fetchone()
            is not None
        )

    def close(self):
        """Closes the store's database."""
        self.cursor.close()
        self.connection.close()

    def __enter__(self):
        """Enables the use of the "with" statement.

        Returns:
            Display: A store context that calls :py:meth:`close` on exit.
        """
        return self

    def __exit__(
        self,
        type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        """Enables the use of the "with" statement.

        Args:
            type (typing.Optional[typing.Type[BaseException]]): None if the context exits without an exception, and the raised exception's class otherwise.
            value (typing.Optional[BaseException]): None if the context exits without an exception, and the raised exception otherwise.
            traceback (typing.Optional[types.TracebackType]): None if the context exits without an exception, and the raised exception's traceback otherwise.
        """
        self.close()

    def __getstate__(self):
        return pathlib.Path(self.path)

    def __setstate__(self, state: pathlib.Path):
        self.__init__(path=state)


class Store(ReadOnlyStore):
    """Stores the IDs of processed tasks.

    Args:
        path (typing.Union[str, os.PathLike]): Path of the SQLite database file with extension ".db".
        commit_maximum_delay (float, optional): How often changes are commited to the disk, in seconds. Defaults to 0.1.
        commit_maximum_inserts (int, optional): Maximum number of changes before commiting changes to the disk. Defaults to 100.
    """

    class Reset:
        """Message requesting a reset of the database (existing entries are dropped)."""

        pass

    class Commit:
        """Message requesting a commit (changes are immediately persited to the disk)."""

        pass

    def target(self):
        """Worker thread implementation."""
        thread_connection = sqlite3.connect(self.path)
        cursor = thread_connection.cursor()
        while self.running:
            commit = False
            for _ in range(0, self.commit_maximum_inserts):
                try:
                    message = self.queue.popleft()
                    if isinstance(message, Store.Reset):
                        rows = [
                            row for row in cursor.execute("pragma table_info(complete)")
                        ]
                        if len(rows) != 1 or rows[0] != (0, "id", "TEXT", 1, None, 1):
                            raise Exception(
                                'the table "complete" does not have the expected format'
                            )
                        cursor.executescript(
                            "drop table if exists complete; create table complete (id text primary key) without rowid;"
                        )
                        thread_connection.commit()
                        commit = False
                    elif isinstance(message, Store.Commit):
                        thread_connection.commit()
                        self.commit_barrier.wait()
                        commit = False
                    else:
                        cursor.execute(
                            "insert or ignore into complete values (?)", (message,)
                        )
                        commit = True
                except IndexError:
                    break
            if commit:
                thread_connection.commit()
            else:
                time.sleep(self.commit_maximum_delay)
        cursor.close()
        thread_connection.close()

    def __init__(
        self,
        path: typing.Union[str, os.PathLike],
        commit_maximum_delay: float = 0.1,
        commit_maximum_inserts: int = 100,
    ):
        super().__init__(path=path)
        self.commit_maximum_delay = commit_maximum_delay
        self.commit_maximum_inserts = commit_maximum_inserts
        self.queue: collections.deque[
            typing.Union[str, Store.Reset, Store.Commit]
        ] = collections.deque()
        self.running = True
        self.thread = threading.Thread(target=self.target, daemon=True)
        self.thread.start()
        self.commit_barrier = threading.Barrier(2)

    def add(self, id: str):
        """Adds a row to the database.

        The action is ignored if the entry is already in the database.

        Args:
            id (str): Entry to store in the database.
        """
        self.queue.append(id)

    def reset(self):
        """Drops all entries from the database."""
        self.queue.append(Store.Reset())

    def commit(self):
        """Immediately persists changes to the disk."""
        self.queue.append(Store.Commit())
        self.commit_barrier.wait()

    def close(self):
        """Closes the store's database."""
        self.running = False
        self.thread.join()
        super().close()

    def __getstate__(self):  # type: ignore
        return (
            pathlib.Path(self.path),
            self.commit_maximum_delay,
            self.commit_maximum_inserts,
        )

    def __setstate__(self, state: typing.Tuple[pathlib.Path, int, int]):  # type: ignore
        self.__init__(
            path=state[0],
            commit_maximum_delay=state[1],
            commit_maximum_inserts=state[2],
        )


if __name__ == "__main__":
    import time

    with Store("progress.db") as store:
        store.reset()
        test_ids = [
            f"nmnist/train/0/{name}"
            for name in os.listdir(
                "/Users/alex/drive/icns/undr_user/datasets/nmnist/train/0"
            )
        ]
        print(f"{len(test_ids)=}")
        begin = time.monotonic()
        for test_id in test_ids:
            store.add(test_id)
        store.commit()
        end = time.monotonic()
        print(f"{end - begin=} s ({(end - begin) / len(test_ids) * 1e3} ms/add)")
        begin = time.monotonic()
        for test_id in test_ids:
            assert test_id in store
        end = time.monotonic()
        print(f"{end - begin=} s ({(end - begin) / len(test_ids) * 1e3} ms/in)")
